/*                        Copyright (c) 1993 Bellcore
 *                            All Rights Reserved
 *       Permission is granted to copy or use this program, EXCEPT that it
 *       may not be sold for profit, the copyright notice must be reproduced
 *       on copies, and credit should be given to Bellcore where it is due.
 *       BELLCORE MAKES NO WARRANTY AND ACCEPTS NO LIABILITY FOR THIS PROGRAM.
 */

/* stim.c  -  MICRO/C-Z8 program for the STIM (Stephen Uhler 6/93)
 *
 * $Header: c:/mcc/tests/RCS/sau1.c 1.6 93/07/04 06:23:02 sau Exp Locker: sau $
 *
 * This is the Wendi STIM version, it has:
 * - Squelch detection (1811, 1840)
 * - ringing debounce (1817, 1840)
 * - extension pickup detection
 *
 * Strategy:
 *
 * Poll ports (including UART), waiting for an event.  Process events as
 * received.  the "counter" is the global sense of time, incrementing
 * every 100th second, and resetting after 0xFF.
 * Remember, the z8 has only 239 bytes of memory, so the size
 * of each stack frame is to be minimized
 *
 * This code is compiled with the MCC z8 'C' compiler
 */

#define VERSION "STIM-3.2"

/* define program stack memory map */

#define INTERNAL	  0x04		/* Internal stack select */
#define EXTERNAL	  0x00		/* External stack select */
#define STACK_MAP	INTERNAL		/* select stack map */

/*
 * define the state bits.  We need to do something anytime one of
 * these changes. Bits 2,4, and 8 use the port-3 names and values
 */

#define TEST			0x01	/* unused - for testing only */
#define SQ_ENABLE		0x10	/* Squelch detection enabled */
#define BLINK			0x20	/* blink fast/slow */
#define SENT_RING		0x40	/* sent "R" for ring */
#define UN_RING		0x80	/* ready to send "r" */

/*
 * port 3 bits. 0-3 are inputs, 4-7 are outputs.  0 and 7 are used
 * for the serial port
 */

#define NOT_RING		0x02	/* 0 is ringing */
#define DTMF			0x04	/* valid DTMF tone */
#define LOOP			0x08	/* loop current when true */
#define P3_IN			0x0E	/* 3 in-bits */

#define NO_DCD			0x10	/* DCD on / off  */
#define OFF_HOOK		0x20	/* on / off hook */
#define NC_0			0x40	/* unused */

/* port 2 bits */

#define TONE_MASK		0xF	/* DTMF input/output code */
#define SQUELCH		0x10	/* squelch detect */
#define NC_1			0x20	/* unused */
#define LED				0x40	/* signal light */
#define NC_2			0x80	/* unused */

/* Misc defines */

#define IS_CHAR			(IRQ&8)			/* a character is ready to read */
#define COUNTER			(0xFF)			/* counter range */

/* timing parameters.  All times are in 100ths of seconds */

/*
 * On some DAA's, the RING line toggles with the ringing frequency
 * Wait this long after ringing stopped to debounce the ring signal
 */

#define RING_DEBOUNCE	15

/*
 * pause after enabling DCD before sending 1st character out the serial port
 * to give the host time to open the port
 */

#define WAIT_OPEN			5

/*
 * The loop relay "sees" ringing before the "RING" output of the DAA
 * wait long enough after seeing a loop change to insure its not a ring
 * we haven't seen yet.
 */

#define RING_TIME			35

/*
 * wait for loop relay to debounce after changing an on-hook condition
 */

#define LOOP_DEBOUNCE	10

/*
 * When we force an off-hook, we need to wait for the loop relay to
 * activate to insure loop current is actually established.  Wait this
 * long between checks of the loop relay state.
 */

#define LOOP_WAIT			5

/*
 * Flash-hook on-hook holding time
 */

#define FLASH_WAIT		50

/* global variables */

char last_p2;		/* previous value of port_2 */
char last_p3;		/* previous value of port_3 */
char port3;			/* cached value of port3 */
char port2;			/* cached value of port2 */
char state;			/* current state flags */
char c;				/* input character or event mask */
char i;				/* temporary */
int ring_count;	/* timer for ring debounce */
int count;			/* 1/100 second counter */

main()
	{

	/* Canonical Z8691 Processor Initialization */

	P01M = 0xF3 | STACK_MAP;
   P2M = 0x00;			  /* set Port 2 output */
   P3M = 0x51;			  /* set activate Port 2 pull-ups, */

	/* STIM specific initialization */

	init();

	/* main program loop */
	
	while (1) {
		port3 = P3;									/* cache port inputs */
		port2 = P2;
		c = (port3&P3_IN&state)^last_p3;		/* changed event mask */
		if (c)
			do_event();
		if ((last_p2 ^ port2) & SQUELCH) {	/* check squelch */
			if (state&SQ_ENABLE)
				cputs(port2&SQUELCH ? "S\n": "s\n");
			last_p2 = port2;
			}
		if (IS_CHAR) {								/* got a character */
			c = _getkey();
			do_char();
			}
		}
	}

/* process an input event */

do_event()
	{

	/* check for start of ringing */

	if (c&NOT_RING&~port3  && !(state&SENT_RING)) {
		P3 = port3 & ~NO_DCD;					/* assert DCD */
		sleep(WAIT_OPEN);
		cputs("R\n");
		state = state | SENT_RING;
		}

	/* check for end of ringing (could be a false alarm) */

	if ((c&NOT_RING&port3) && (state&SENT_RING)) {
		ring_count = (count + RING_DEBOUNCE)&COUNTER;
		}	

	/* verified end of ringing (UN_RING is set in timer interrupt routine) */

	if (c&UN_RING)	{
		cputs("r\n");
		last_p3 = last_p3 & ~UN_RING;
		P2 = P2 & ~LED;
		state = state & ~SENT_RING;
		}

	/* got a DTMF tone (ignore tone-off conditions) */

	if (c&DTMF&port3) {
		putchar("D1234567890*#ABC"[port2&TONE_MASK]);
		putchar('\n');
		}
		
	/* got a disconnect while on-hook */

	if (c&LOOP&~port3 && port3&OFF_HOOK) {
		sleep(2);					/* wait 20-30 ms */
		if (P3&LOOP) return;		/* just noise */
		else {						/* disconnect: go on hook */
			cputs("d\n");
			P3 = P3 & ~OFF_HOOK;
			sleep(LOOP_DEBOUNCE);
			state = state | NOT_RING;
			}
		}

	/* on-hook loop change - extension phone detect */

	else if (c&LOOP && !(port3&OFF_HOOK) && !(state&SENT_RING)) {
		if (port3&LOOP) {
			sleep(RING_TIME);
			if (~P3 & NOT_RING)	/* really ringing started! */
				return;
			P3 = port3 & ~NO_DCD;
			sleep(WAIT_OPEN);
			cputs("H\n");
			sleep(LOOP_DEBOUNCE);
			}
		else
			cputs("h\n");
		}

	/* reset event mask */

	last_p3 = port3 & P3_IN & state;
	}

/* process a character */

do_char()
	{
	i = 'N';			/* default character to print */

	switch (c) {
		case 'h':								/* go on hook */
			if (port3&OFF_HOOK) {
				P3 = P3 & ~OFF_HOOK;
				state = state | NOT_RING;
				sleep(LOOP_DEBOUNCE);
				i = 'y';
				}
			break;
		case 'H':								/* go off hook */
			if (!(port3&OFF_HOOK)) {
				P3 = port3 | OFF_HOOK;
				state = state & ~NOT_RING;
				last_p3 = last_p3 & ~NOT_RING;

				/* wait for loop current */

				for(c=0;c<30;c++) {			/* wait up to 1.5 sec for loop */
					if (P3&LOOP) break;
					sleep(LOOP_WAIT);
					}

				if (P3&LOOP) {					/* went off hook */
					i = 'Y';
					sleep(LOOP_DEBOUNCE);	/* wait for loop relay to debounce */
					P2 = port2 | LED;
					}
				else {							/* no loop current */
					P3 = port3 & ~OFF_HOOK;
					}
				}
			break;
		case 'S':								/* enable squelch detection */
			if (!(state&SQ_ENABLE)) {
				state = state | SQ_ENABLE;
				i = 'Y';
				}
			break;
		case 's':								/* disable squelch detection */
			if (state&SQ_ENABLE) {
				state = state & ~SQ_ENABLE;
				i = 'Y';
				}
			break;
		case 'E':								/* enable touch tone detection */
			if (!(state&DTMF)) {
				state = state | DTMF;
				i = 'Y';
				}
			break;
		case 'e':								/* disable touch tone detection */
			if (state&DTMF) {
				state = state & ~DTMF;
				i = 'Y';
				}
			break;
		case 'D':								/* assert DCD */
			if (port3&NO_DCD) {
				P3 = P3 & ~NO_DCD;
				i = 'Y';
				}
			break;
		case 'd':								/* de-assert DCD */
			if (!(port3&NO_DCD)) {
				P3 = port3 | NO_DCD;
				i = 'Y';
				}
			break;
		case 'F':								/* flash hook */
			if (port3&OFF_HOOK) {
				P3 = port3 & ~OFF_HOOK;
				sleep(FLASH_WAIT);
				P3 = port3 | OFF_HOOK;
				sleep(LOOP_DEBOUNCE);
				i = 'Y';
				}
			break;

		case 'V':								/* print version info */
			version();
			i = 0;
			break;
		case 'X':								/* sync */
			putchar('X');
			i = 0;
			break;
		case 'L':								/* enable loop detection */
			if (!(state&LOOP)) {
				state = state | LOOP;
				i = 'Y';
				}
			break;
		case 'l':								/* disable loop detection */
			if (state&LOOP) {
				state = state & ~LOOP;
				i = 'Y';
				}
			break;
		case 'C':								/* set fast blink */
			if (!(state&BLINK)) {
				state = state | BLINK;
				i = 'Y';
				}
			break;
		case 'c':								/* set slow blink */
			if (state&BLINK) {
				state = state & ~BLINK;
				i = 'Y';
				}
			break;
		case '?':								/* return status ADHRELS. */
			putchar('A');
			putchar(port3&NO_DCD?'d':'D');
			putchar(port3&OFF_HOOK?'H':'h');
			putchar(port3&NOT_RING?'r':'R');
			putchar(state&DTMF?'E':'e');
			putchar(state&LOOP?'L':'l');
			putchar(state&SQ_ENABLE?'S':'s');
			i = '.';
			break;
		case 'R':								/* reset */
			init();
			i = 'Y';
			break;
		}
	if (i) putchar(i);
	putchar('\n');
	}

/* put a character string */

cputs(ptr)
char *ptr;
	{
	while (*ptr) putchar(*ptr++);
	}

/* initialize stim to the default state */

init()
	{

	/* setup serial port control hardware (9600 BAUD @7.3728 MHz) */

	PRE0 = (3<<2) | 1;	/* set timer0 prescaler  */
	T0 = 02;					/* set timer0 count      */
	TMR = 0x03;				/* load/start timer0     */
	IRQ = 0x10;				/* set serial transmit (irq4) ready */

	/* setup the interval timer
	 * interval = 8/clock * T1 * PRE1
    * clock = 7,372,800 HZ
    * 7372800/8 * 64 * 144 = .01 sec.
    */

	PRE1 = (63<<2) | 0x3;	/* set pre-scaler */
	T1 = 144;					/* set counter */
	TMR = TMR | 0x0C;			/* enable counter */
	IPR = 0x01;					/* set interrupt priority */
	IMR = 0xa0;					/* enable Timer 1 interrupts */

	P2M = 0x1F;					/* port i/o modes */
	P3M = 0x41;					/* port 3 modes */
	P2 = LED;					/* initialize port2 */
	P3 = NO_DCD;				/* initialize port3 */

	sleep(LOOP_DEBOUNCE);	/* wait for loop relay to debounce */
	last_p3 = P3&P3_IN;		/* initialize current P3 state */
	last_p2 = P2;				/* initial value of squelch */
	count = 0;					/* initialize clock counter */
	state = NOT_RING;			/* look for ringing */
	ring_count = COUNTER+1;	/* impossible counter value */
	}

/* This runs 100 times per second, at interrupt time */

timer() interrupt 5 using 2
	{
	count = (count+1)&COUNTER;
	
	if (count == ring_count) {				/* ringing off */
		last_p3 = last_p3 | UN_RING;		/* will cause 'r' next do_event */
		ring_count = COUNTER+1;
		return;
		}

	/* if off-hook leave the LED on */

	if (port3&OFF_HOOK)
		return;

	/* "ring" with LED */

	if (state&SENT_RING)
		switch(count&0xF) {
			case 0:  P2=port2|LED; break;
			case 4:  P2=port2&~LED; break;
			}

	/* LED heart-beat. The trigger values determine the blinking sequence */

	else switch(count) {
		case 0: case 128:						/* turn LED on */
			P2 = port2 | LED; break;
		case 20: case 148:					/* turn LED off */
			P2 = port2 & ~LED; break;
		case 65: case 85:						/* add extra flash for BLINK */
			if (state&BLINK) P2 = port2 ^ LED; break;
		}
	}

/* sleep in 10ms (.01 sec) intervals */

sleep(n)
int n;
	{
	n = (n + count)&COUNTER;
	while(n != count);						/* wait for timer to catch up */
	}

/* print version info */

version()
	{
	cputs(VERSION);
	cputs(" (c) 1992,1993 Bellcore (by S. A. Uhler)");
	}
