/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* audio interface subs (TAKE II) S. A. Uhler */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <sun/audioio.h>
#include <sys/signal.h>
#include <stdio.h>
#include "audio_lib.h"

#define AUDIO			"/dev/audio"

#define dprintf		if (debug>0) fprintf
#define Dprintf		if (debug>1) fprintf
#define dputc			if (debug) putc

/* these can be overridden by an application that knows about them */

int record_gain = R_GAIN;
int play_gain = P_GAIN;

static unsigned int wrote = 0;      /* total bytes written */

/* set sparc audio parameters */

static struct audio_info save_info; 
static int debug = 0;

/* open the audio device */

int
audio_open(how)
int how;
	{
	char *getenv();
	int fd;

	dprintf(stderr,"a: opening audio\n");

	fd = open(AUDIO,how);
	if (fd>0) {
		record_stop(fd);
		fcntl(fd,F_SETFD,FD_CLOEXEC);
		audio_set(fd,record_gain,play_gain);
		}
	dprintf(stderr,"a: opening audio fd=%d\n",fd);
	return(fd);
	}

/* close the audio stream */

int
audio_close(fd)
int fd;
	{
	dprintf(stderr,"a: closing audio %d\n",fd);
	if (fd >0) {
		audio_flush(fd,2);
		audio_restore(fd);
		return(close(fd));
		}
	else
		return(-1);
	}

static int last_rec, last_play;		/* saved values */

int 
audio_set(fd,rec,play)
int fd;		/* fd for audio control */
int rec;		/* record level */
int play;	/* playback level */
	{
	struct audio_info info; 
	static int first = 1;
	int n;

	if (fd < 0) return(0);

	ioctl(fd,AUDIO_GETINFO,&info);
	
	if (first) {
		save_info = info;
		first = 0;
		}

	last_play = info.play.gain;
	last_rec = info.record.gain;

	if (play > 0) {
		info.play.gain = play;
		}
	if (rec > 0) {
		info.record.gain = rec;
		}
	info.play.port =  AUDIO_HEADPHONE;

	n = ioctl(fd,AUDIO_SETINFO,&info);
	dprintf(stderr,"a: setting audio gain to %d/255 %d/255 %s\n",
			rec,play,n?"FAILED":"");
	return(n==0);
	}

/* restore previous audio values */

int 
audio_pop(fd)
int fd;		/* fd for audio control */
	{
	struct audio_info info; 

	if (fd < 0) return(0);

	Dprintf(stderr,"Popping %d (%d,%d)\n",fd,last_rec, last_play);
	ioctl(fd,AUDIO_GETINFO,&info);
	info.record.gain = last_rec;
	info.play.gain = last_play;
	if (!ioctl(fd,AUDIO_SETINFO,&info))
		perror("IOCTL audio_pop");
	return(1);
	}

/* wait for audio to drain */

int
audio_drain(fd)
int fd;
	{
	int n;

	if (fd < 0)
		return(-1);

	dprintf(stderr,"a: draining audio\n");
	if (n=(ioctl(fd,AUDIO_DRAIN,0) != 0)) 
		perror("ioctl AUDIO_DRAIN");
	return(!n);
	}
	
int
audio_restore(fd)
int fd;
	{
	struct audio_info info; 

	if (fd < 0)
		return(-1);

	dprintf(stderr,"a: Restoring audio parameters\n");
	ioctl(fd,AUDIO_GETINFO,&info);
	info.play.gain = save_info.play.gain;
	info.record.gain = save_info.record.gain;
	info.play.port = save_info.play.port;
	return(ioctl(fd,AUDIO_SETINFO,&info)==0);
	}


int
audio_mode(fd,what,how)
int fd;
int what;						/* 0 = record, 1 = play */
int how;							/* FALSE = PAUSE, TRUE = play */
	{
	struct audio_info info; 
	int n;

	if (fd < 0)
		return(-1);

	ioctl(fd,AUDIO_GETINFO,&info);
	if (what)
		info.play.pause = how;
	else
		info.record.pause = how;
	n = ioctl(fd,AUDIO_SETINFO,&info);
	Dprintf(stderr,"A: AUDIO PARAMETERS: %s %s %s\n",
			what?"play":"record", how?"un-pause":"pause",n?"FAILED":"");
	return(n);
	}

/* flush the audio output (thus stopping it) (return samples written) */
/* I think this is broken (sau 5.92) */

int
audio_flush(fd,how)
int fd;
int how;			/* flush input/output/ or both */
   {
	struct audio_info info; 
	int result = 0;
	int n;		/* temp for debugging */

	if (fd < 0)
		return(-1);

   n = ioctl(fd,AUDIO_GETINFO,&info);
	if (how) {
		result = info.play.samples;
		wrote = info.play.samples = 0;
		info.play.pause = 1;
		}
	else {
		result = info.record.samples;
		wrote = info.record.samples = 0;
		info.record.pause = 1;
		}
   n |= ioctl(fd,I_FLUSH,how+1);
   n |= ioctl(fd,AUDIO_SETINFO,&info);
	if (how) {
		info.play.pause = 0;
		n |= ioctl(fd,AUDIO_SETINFO,&info);
		}
	Dprintf(stderr,"A: Flushing audio %s\n",n?"IOCTL FAILED":"");
   return(result);
   }

/* reset audio write counter (return previous value) */

int
audio_reset(fd)
int fd;
   {
	struct audio_info info; 
	int result;
	int n;

	if (fd < 0)
		return(-1);

   n = ioctl(fd,AUDIO_GETINFO,&info);
   result = info.play.samples;
   wrote = info.play.samples = 0;
   n |=ioctl(fd,AUDIO_SETINFO,&info);
	Dprintf(stderr,"A: resetting audio counter %s\n",n?"IOCTL FAILED":"");
   return(result);
   }

/* audio read */

int
audio_read(fd,buff,bytes)
int fd;
char *buff;
int bytes;
	{
	int count;
	static int log_file = 0;
	char *getenv();

	if (fd < 0)
		return(-1);

	if (log_file == 0)
		log_file = open(getenv("LOG_RECORD"),O_WRONLY|O_CREAT,0600);
		
	count = read(fd,buff,bytes);
	if (log_file > 0)
		write(log_file,buff,count);
	return(count);
	}

/* write audio */

int
audio_write(fd,buff,bytes)
int fd;     /* audio fd */
char *buff;    /* what to write */
int bytes;
   {
   int count;

	if (fd < 0)
		return(-1);

	Dprintf(stderr,"A: writing %d ...",bytes); if (debug>1) fflush(stderr);
   count = write(fd,buff,bytes);
   if (count > 0)
      wrote += count;
	Dprintf(stderr,"A: wrote %d of %d  (total %d) bytes of audio\n",
			count,bytes,wrote);
   return(count);
   }

/* adjust audio write counter */

int
audio_wrote(n)
int n;
	{
	wrote += n;
	return(wrote);
	}

/* how many bytes are in the audio Q */

int
audio_writeq(fd)
int fd;
   {
	int n;
	struct audio_info info; 

	if (fd < 0)
		return(0);

   n = ioctl(fd,AUDIO_GETINFO,&info);
	Dprintf(stderr,"A: writeQ is %d %s\n",
			wrote-info.play.samples,n?"IOCTL FAILED":"");
   return(wrote - info.play.samples);
   }

void
debug_audio(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"audio_debug is %d\n",n);
	}
