/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* simple database routines (S. A. Uhler) */

#include <stdio.h>
#include <string.h>
#include "db.h"

#define dprintf		if (debug) fprintf
#define Dprintf		if (debug>1) fprintf

struct db {
	struct data *top;			/* pointer to attr names above 1st row */
	struct data *current;	/* pntr to current item */
	char *name;					/* name of file */
	int state;					/* unused state flags */
	int count;					/* number of lines */
	int fields;					/* number of fields */
	};

struct data {
	struct data *next;		/* pntr to next row */
	char *data;					/* data in this row */
	char selected;				/* true if item is selected */
	};

char *malloc(), *getenv(), *index();
static char *save_line();
static char *isolate_item();
static int find_item();
static int free_data();

static int debug = 0;

char line[1024];				/* spadce for longest line */

/* open a database, read into memory, return database cookie */

char *
db_open(name)
char *name;
	{
	struct db *db;
	struct data *data, *prev=NULL;
	FILE *file = fopen(name,"r");
	int count = 0;

	dprintf(stderr,"d: opening database %s\n",name);
	if (!file) {
		dprintf(stderr,"d: Can't open %s\n",name);
		return(NULL);
		}

	db = (struct db *) malloc(sizeof(struct db));
	db->name = save_line(name);
	db->state = 0;
	db->current = NULL;

	while(fgets(line,sizeof(line),file)) {
		if (count==0 && *line=='#') 
			continue;
		Dprintf(stderr,"D: (file %s) got line %s",name,line);
		data = (struct data *) malloc(sizeof(struct data));
		data->data = save_line(line);
		data->selected = 1;
		data->next = NULL;
		if (prev) 
			prev->next = data;
		else
			db->top = data;
		prev = data;
		count++;
		}
	db->count = count;
	db->fields = count_delim(db->top,DELIM);
	dprintf(stderr,"d: opened %s, %d items\n",name,count);
	return((char *) db);
	}

/* write oput the database */

static char buff[100];
static char buff2[100];

int
db_write(s,name)
char *s;		/* cookie */
char *name;	/* name of file, NULL=>same name */
	{
	struct db *db = (struct db *) s;
	struct data *data;
	char *path;
	FILE *file;


	if (name)
		path = name;
	else {
		sprintf(buff,"%s.%d",db->name,getpid());
		path=buff;
		}

	dprintf(stderr,"d: writing %s as %s\n",db->name,path);

	if ((file = fopen(path,"w")) == NULL){
		dprintf(stderr,"d: Save can't open %s\n",path);
		return(0);
		}

	for(data=db->top;data;data=data->next) {
		if (data->selected)
			fwrite(data->data,1,strlen(data->data),file);
		}

	fclose(file);
	if (!name) {		/* do commit stuff */
		sprintf(buff2,"%s.old",db->name);
		rename(db->name,buff2);
		link(path,db->name);
		unlink(buff2);
		unlink(path);
		}
	return(0);
	}

/* close a database, free up space */

int
db_close(s)
char *s;		/* cookie */
	{
	struct db *db = (struct db *) s;
	struct data *next,*data;

	if (!db)
		return(-1);
	dprintf(stderr,"d: closing %s\n",db->name);
	for(data=db->top;data;data=next) {
		next = data->next;
		free_data(data);
		}
	free(db->name);
	free(db);
	return(0);
	}

/* goto to the top */

int
db_top(s)
char *s;
	{
	struct db *db = (struct db *) s;

	dprintf(stderr,"d: Setting %s to top\n",db->name);
	db->current = db->top;
	return(1);
	}

/* update a value in the current item */

int
db_update(s,item,value)
char *s;
char *item;				/* item to update */
char *value;				/* value to update with */
	{
	struct db *db = (struct db *) s;
	int n =  find_item(db,item);
	struct data *data = db->current;		/* pointer to current value */
	char *current;						/* current value to be updated */
	char *end;							/* next item */
	int len = strlen(value);		/* length of new value */
	int olen;							/* length of old value */

	if (n<0) {
		dprintf(stderr,"d: no item %s to update\n",item);
		return(0);
		}
	if (!data) {
		dprintf(stderr,"d: no current value in %s to update\n",db->name);
		return(0);
		}

	current = isolate_item(data->data,n);	/* find item */
	end = isolate_item(data->data,n+1);	  /* beginning of next item */
	olen = end-current-1;
	if (olen  == len) {/* in-place update */
		bcopy(value,current,len);
		Dprintf(stderr,"D:  INPLACE Updating %s to %s\n",item,value);
		}
	else {			/* do it the hard way */
		char *new = malloc(strlen(data->data) + len - olen + 1);
		int mark = current-data->data;		/* len in frot of item */

		Dprintf(stderr,"D:  Updating %s to %s\n",item,value);
		bcopy(data->data,new,mark);		/* front */
		bcopy(value,new+mark,len);					/* new value */
		strcpy(new+mark+len,current+olen);		/* end */
		free(data->data);
		data->data = new;
		}
	return(1);	
	}

/* fetch a value out of a row */

char *
db_fetch(s,item,value)
char *s;
char *item;			/* which item to fetch */
char *value;		/* put the value here! (malloc space if null) */
	{
	struct db *db = (struct db *) s;
	int n =  find_item(db,item);
	struct data *data = db->current;		/* pointer to current value */
	char *current;								/* pointer to current item */
	char *end;									/* end of item */

	if (n<0) {
		dprintf(stderr,"d: no item %s to fetch\n",item);
		return(NULL);
		}

	if (!data) {
		dprintf(stderr,"d: no current value in %s to update\n",db->name);
		return(NULL);
		}
	
	current = isolate_item(data->data,n);	/* find item */
	end = isolate_item(data->data,n+1);	  /* beginning of next item */

	if (!value)
		value = malloc(end-current);
	bcopy(current,value,end-current-1);
	value[end-current-1] = '\0';
	Dprintf(stderr,"D: fetch of [%s]: returning [%s]\n",item,value);
	return(value);
	}

/* delete the current line make next item current */

int
db_delete(s)
char *s;
	{
	struct db *db = (struct db *) s;
	struct data *current;
	struct data *prev;

	dprintf(stderr,"d: deleting in %s\n", db->name);

	if (!(current=db->current)) {
		dprintf(stderr,"d: No current item to delete\n");
		return(0);
		}

	for(prev=db->top;prev && prev->next != current;prev=prev->next);

	if (prev->next) {
		prev->next = db->current->next;
		db->count--;
		db_next(s);	
		free_data(current);	
		return(1);
		}
	else {
		dprintf(stderr,"d: Oops, cant delete?? (%s)\n",db->name);
		return(0);
		}
	}
		
/* insert a new line after the current one, make new item current */

int
db_insert(s,row,sel)
char *s;
char *row;		/* line to insert - NO CHECKING */
int sel;			/* if true, set to selected */
	{
	struct db *db = (struct db *) s;
	struct data *data;

	dprintf(stderr,"d: (%s) Inserting %s\n",
			db->name,row);

	if (count_delim(row,DELIM) != db->fields) {
		dprintf(stderr,"d: Insert failed, wrong # of fields\n");
		return(-1);
		}

	data = (struct data *) malloc(sizeof(struct data));
	if (!db->current)
		db->current=db->top;

	data->next = db->current->next;
	data->data = save_line(row);
	data->selected = sel;
	db->current->next = data;
	db->current = data;
	db->count++;
	return(1);
	}
		
/* save a database position */

char *
db_mark(s)
char *s;
	{
	struct db *db = (struct db *) s;

	dprintf(stderr,"d-mark:  (%.10s...)\n",db->current->data);
	return((char *)db->current);
	}

/* make current item the one saved from a db_save call */

int
db_set(s,where)
char *s;
char *where;
	{
	struct db *db = (struct db *) s;
	struct data *set = (struct data *) where;

	if (db->current == set)
		return(1);

	db->current = db->top;
	for(;db->current && db->current != set;db->current=db->current->next);
	return(db->current == set);
	}

/* make next item current, return pointer to row */

char *
db_next(s)
char *s;
	{
	struct db *db = (struct db *) s;

	if (db->current)
		db->current = db->current->next;

	while (db->current && db->current->selected==0) {
		db->current = db->current->next;
		}

	if (db->current) {
		dprintf(stderr,"d-next:  (%.10s...)\n",db->current->data);
		return(db->current->data);
		}
	else {
		dprintf(stderr,"d-next:  NULL\n");
		return(NULL);
		}
	}
		
/* select items from a database */

int
db_select(s,item,expr,how)
char *s;		/* database cookie */
char *item;	/* name of item */
char *expr;	/* what to select on */
int how;		/* SET, OR, or AND */
	{
	struct db *db = (struct db *) s;		/* pointer to db */
	struct data *data = db->top->next;	/* current item */
	char *current;								/* current item value */
	int n =  find_item(db,item);
	int cmp=atoi(expr);						/* incoming value for comparisons */
	int len = strlen(expr);	
	int result;									/* result comparison */
	int ok;										/* selection passed */
	int count=0;
	char tmp[32];

	if (n<0) {
		dprintf(stderr,"db: item %s not found in %s\n",item,db->name);
		return(0);
		}

	if (how&SEL_REG && re_comp(expr) != NULL) {
		dprintf(stderr,"db (%s): invalid re-exp %s\n",db->name,expr);
		return(0);
		}

	dprintf(stderr,"d: Selecting item=[%s] expr=[%s] from %s [",
			item,expr,db->name);
	if (debug)
		print_sel(how);
	dprintf(stderr,"]\n");
	for(;data;data=data->next) {
		Dprintf(stderr,"Select: working on %s",data->data);
		if ((how&SEL_OP)==SEL_OP) {		/* SET or CLEAR (special case ) */
			if (how&SEL_OR) {
				Dprintf(stderr,"  SET\n");
				data->selected=1;
				count++;
				}
			else {
				data->selected=0;
				Dprintf(stderr,"  CLEAR\n");
				}
			}
		else {								/* evaluate the expression */
			if (data->selected && (how&SEL_OR))  {	/* already selected */
				count++;
				Dprintf(stderr,"  Already Selected %d (using OR) \n",count);
				continue;
				}
			if (!data->selected && (how&SEL_AND)) {	/* already UNselected */
				Dprintf(stderr,"  Already DEselected (using AND)\n");
				continue;
				}

			current = isolate_item(data->data,n);	/* find item */
			if (how&SEL_TT)  {							/* compare as touch tones */
				cnv2tt(current,tmp);
				current = tmp;
				Dprintf(stderr,"  using TT: %s -> %s\n",current,tmp);
				}
			Dprintf(stderr,"   Comparing %s - %.*s ",expr,len,current);

			if (how&SEL_REG)
				result = !(re_exec(current) == 1);
			else
				result = how&SEL_NUM?cmp-atoi(current):strncmp(expr,current,len);

			ok = 0;
			if (how&SEL_EQ && result==0) ok=1;
			if (how&SEL_LS && result<0)  ok=1;
			if (how&SEL_GR && result>0)  ok=1;

			data->selected = ok;
			if (data->selected)
				count++;
			Dprintf(stderr,"result is %d, selection = %s, %d\n",
						result,ok?"TRUE":"FALSE",count);
			}
		}
	dprintf(stderr,"d: db_select selected %d\n",count);
	return(count);
	}

/* return index # of item in attribute list */

static int
find_item(db,item)
struct db *db;
char *item;
	{
	char *items = db->top->data;
	int n=0;
	int len=strlen(item);

	Dprintf(stderr,"D: Looking for item [%s] (%d) in %s\n",item,len,items);
	for(;items;n++) {
		Dprintf(stderr,"%s ==? %s %d\n",items,item,len);
		if (strncmp(items,item,len)==0) {
			dprintf(stderr,"d: (%s) Item [%s] is index %d\n",db->name,item,n);
			return(n);
			}
		items = index(items,DELIM);
		items?items++:NULL;
		}
	dprintf(stderr,"D: cant find %s\n",item);
	return(-1);
	}

/* isolate an item, base on its index */

static char *
isolate_item(row,indx)
char *row;			/* line to isolate */
int indx;			/* which item */
	{
	while(indx-->0)
		row = index(row,DELIM) + 1;
	Dprintf(stderr,"     isolating %.6s...\n",row);
	return(row);
	}
		

/* free an item */

static int
free_data(data)
struct data *data;
	{
	if (data->data)
		free(data->data);
	return(free(data));
	}

/* allocate space for and save a line (whimpy version) */

static char *
save_line(s)
char *s;		/* line to save */
	{
	Dprintf(stderr,"D: saving [%s]\n",s);
	return(strcpy(malloc(strlen(s)+1),s));
	}

/* convert an arbitrary character string into touch tones */

char cnvt_tt[] = {
	2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,0,7,7,8,8,8,9,9,9,0
	};

int
cnv2tt(s,out)
char *s;							/* the source */
char *out;						/* where to put the result */
	{
	char *save;
	char c;
	int n;

	if (!out)
		out = s;
	save = out;

	Dprintf(stderr,"Converting [%s] to ",s); fflush(stderr);
	while(c = *s) {
		if (c>='0' && c<='9')
			*out++ = c;
		else if (c>='A' && c<='Z' && (n=cnvt_tt[*s - 'A']))
				*out++  = '0' + n;
		else if (c>='a' && c<='z' && (n=cnvt_tt[*s - 'a']))
				*out++  = '0' + n;
		else if (c=='#' || c=='*')
			*out++ = c;
		s++;
		}
	*out = '\0';
	Dprintf(stderr,"%s\n",save);
	return(strlen(save));
	}

/* debugging routine to print selection critereon */

char *p_log[] = {"assign","or","and","and/or"};
char *p_eq[] = {"?","==",">",">=","<", "<=","<>","<=>"};

int
print_sel(how)
int how;
	{
	char *foo;

	fprintf(stderr,"%s ",p_log[how&3]);
	fprintf(stderr,"%s ",p_eq[(how>>4)]);
	switch(how&(SEL_NUM|SEL_TT|SEL_REG)) {
		case SEL_NUM: foo="numeric"; break;
		case SEL_TT: foo="touch-tones"; break;
		case SEL_TT|SEL_NUM: foo="numeric-touch-tones"; break;
		case SEL_REG: foo="regular expressions"; break;
		default: foo="alpha"; break;
		}	
	fprintf(stderr,"%s",foo);
	return(1);
	}

/* count the DELIMS in a line */

int
count_delim(row,delim)
char *row;
char delim;
	{
	int n;

	for(n=0;row&&(row=index(row,delim));n++,row++);
	return(n);
	}
		
void
debug_db(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"db_debug in %d\n",n);
	}
