/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* do call progress stuff version 2 (SAU) */

#include <stdio.h>
#include <math.h>
#include "dsp.h"
#include "progress.h"

/* constants & misc */

#define FILTERS	4					/* number of filters  */
#define NQ			(FILTERS+1)		/* number of Q's */
#define MAXQ		4					/* max elements in each Q (power of 2) */
#define INQ			FILTERS			/* name of input Q */
#define SMPMS		8					/* samples per msec */
#define ON			1					/* on, waiting for off */
#define OFF			0					/* off, waiting for on */

/* adjustable parameters */

#define DRAIN		0.1				/* integrator drain (you pick) */
#define THRES1		0.7				/* theshold for ON detection */
#define THRES2		0.5				/* theshold for OFF detection */
#define SAMPLES	800				/* # samples for detection */
#define GAIN		4000.0			/* initial input gain factor */

#define BETWEEN(a,x,b)     ((x)<(a)?a:((x)>(b)?b:x))
#define ABS(x)					((x)<0.0 ? -(x) : (x))

/* band pass filter element */

#define Filter(c,x0,x2,y1,y2) \
	(c)->k*((x0) - (x2)) - ((c)->b1*(y1) + (c)->b2*(y2))

/* Queue management */

#define ADDQ(i,x)		(q[((i)<<2)+cq] = (x))
#define FETCHQ(i,n)	(n+cq >=0 ? q[((i)<<2)+n+cq] : q[((i)<<2)+cq+n+MAXQ])
#define BUMPQ()		(cq = (cq+1)&3)

/* threshold parameters for detectors */

static float thres = THRES1;
static float gain = GAIN;
static float drain = DRAIN;
static int minsamp = SAMPLES;

struct sum {
	float sum;		/* current integrated value */
	int state;		/* current output state */
	int count;	/* current count in "other" state */
	};

/* coefficents for call progress tone filters
 * sampling rate 8khz, filter gain: 10
 */

static struct coef coef[] = {
	/*   k       a0   a1    a2        b1        b2                  */
#ifndef OLD
   0.00391165, 1.0, 0.0, -1.0, -1.91738090, 0.9921767,		/* 350 HZ */
   0.00391165, 1.0, 0.0, -1.0, -1.87440075, 0.9921767,		/* 440 HZ */
   0.00282212, 1.0, 0.0, -1.0, -1.85605439, 0.99623718,		/* 480 HZ */
   0.00364324, 1.0, 0.0, -1.0, -1.76323824, 0.99514235,		/* 620 HZ */
#else
	0.0205884, 1.0, 0.0, -1.0, -1.92226842, 0.99725488,		/* 350 HZ */
	0.0258735, 1.0, 0.0, -1.0, -1.87851569, 0.99655021,		/* 440 HZ */
	0.0282212, 1.0, 0.0, -1.0, -1.85605439, 0.99623718,		/* 480 HZ */
	0.0364324, 1.0, 0.0, -1.0, -1.76323824, 0.99514235,		/* 620 HZ */
#endif
	};

#define dprintf	if(debug)fprintf
static int debug = 0;

/* mu-law / binary conversion stuff */

extern unsigned char  Ub2i[8192];
extern unsigned short I2ub[256];
extern unsigned char  *Sb2i;
extern short I2sb[256];

#define	B13ZERO	0x1000		/* zero offset in unsigned 13-bit tables */

/* state to mode mapping */

static int modes[] = {
	NONE, INVALID, INVALID, DIAL_TONE, INVALID, INVALID, RINGING, INVALID,
	INVALID, INVALID, INVALID, INVALID, BUSY, INVALID, INVALID, INVALID
	};

/* mode names (for debugging) */

static char *names[] = {
	"none", "350", "440", "dial tone",
	"480", "350+480", "ringing", "350+440+480",
	"620", "350+620", "440+620", "350+440+620",
	"busy", "350+480+620", "440+480+620", "350+440+480+620"
	};
	
static float q[NQ*MAXQ];					/* current Q values */
static struct sum sums[FILTERS];			/* place to hold sums */
static int cq = 0;							/* current index into Q's */
static int state = 0;						/* global filter state */
static int first = 1;						/* first time initialization */

/* do call progress monitoring.  Return # of bytes processed */

int
progress(in,count,mask)
register char *in;				/* where to get sound from */
int count;							/* # of bytes of sound info */
int *mask;							/* what to look for (return the reason) */
	{
	register int i;				/* filter number */
	register float x;				/* current scaled sample */
	register float x0,x2;		/* input values */
	register char *end = in + count;
	char *getenv();

	/* initialization */

	if (first) {
		gain = getenv("GAIN") ? atof(getenv("GAIN")) : GAIN;
		drain = getenv("DRAIN") ? atof(getenv("DRAIN")) : DRAIN;
		}
	if (count==0 || first==1) {
		first = state = cq = 0;
		dprintf(stderr,"progress: drain is %f, gain is %f\n",drain,gain);
		for(i=0;i<FILTERS;i++) {
			sums[i].sum = 0.0;
			sums[i].state = OFF;
			sums[i].count = 0;
			}	
		for(i=0;i<NQ*MAXQ;i++)
			q[i] = 0.0;
		if (count==0)
			return(0);
		}

	/* look at the data */

	while (in < end) {
		x = I2sb[(int)*in++]/gain;		/* scale sample to +- 1 */
		ADDQ(INQ,x);					/* feed the pipeline */

		x0 = FETCHQ(INQ,0);			/* use temporaries for loop invarients */
		x2 = FETCHQ(INQ,-2);

		/* do the filters */

		for(i=0;i<FILTERS;i++) {

			x = Filter((coef+i), x0, x2, FETCHQ(i,-1), FETCHQ(i,-2));
			ADDQ(i,x);

			/* compute the thresholds */

			x = ABS(x);
			x += sums[i].sum - drain;
			sums[i].sum = BETWEEN(0.0,x,1.0);
			switch (sums[i].state) {
				case OFF:		/* look for a hit */
					if (x > thres)
						sums[i].count++;
					else
						sums[i].count = 0;
					break;	
				case ON:		/* wait for decay  */
					if (x < THRES2)
						sums[i].count++;
					else
						sums[i].count = 0;
					break;	
				}

			/* got a tone transition */

			if (sums[i].count > minsamp) {
				int mode;

				sums[i].count = 0;
				sums[i].state = 1-sums[i].state;

				if (sums[i].state == ON)
					state |= (1<<i);
				else
					state &= ~(1<<i);

				mode = modes[state];
				dprintf(stderr,"  progress mode %d \"%s\" (%d)\n",
							mode,names[state],state);

				if (mode & *mask) {
					*mask = mode;
					return(in-end+count);
					}
				}
			}
		BUMPQ();
		}
/*
	dprintf(stderr,"progress: (%d) %d%d%d%d %.2f %.2f %.2f %.2f\n",
			*mask, sums[0].state,sums[1].state,sums[2].state,sums[3].state,
			sums[0].sum,sums[1].sum,sums[2].sum,sums[3].sum);
*/
	*mask = 0;
	return(in-end+count);
	}

void
debug_progress(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"progress_debug in %d\n",n);
	}
