/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* misc utility routines that require phone and audio */

#include <stdio.h>
#include <sys/time.h>
#include <sys/filio.h>
#include "audio_lib.h"
#include "phone_lib.h"

#define dprintf      if (debug) fprintf
#define Dprintf      if (debug>1) fprintf
#define dputc        if (debug) putc
#define Min(x,y)		((y)<(x)?(y):(x))

#define MIN_Q			(SPS/4)				/* bytes to leave in audio Q */
#define MAX_PLAY		(SPS*60*60)			/* maximum playable file length */

#define C_IFLUSH		'N'					/* pre-flush phone fd */
#define C_OFLUSH		'X'					/* don't flush audio after tone */
#define C_QUIET		'Q'					/* do quiet detection */
#define C_DIAL			'D'					/* do dial-tone detection */
#define C_TIMEOUT		'T'					/* timeout character */

static int debug = 0;
static char buff[HUNK];		/* audio bufffer */
static char line[10];			/* phone input line */
char *index();	
char *getenv();

/* wait specified number of rings to answer */

int
wait_rings(phone_fd,rings)
int phone_fd;				/* fd of phone */
int rings;					/* # of rings to wait */
	{
	int ring = 0;
	int n;

	set_timeout(150);
	while(ring < rings && !is_timeout()) {
		n = phone_read(phone_fd,line,sizeof(line),0);
		if (n<0)
			return(n);
		switch(*line&0x7f) {
			case 'R':
				alarm(0);
				break;
			case 'r':
				ring++;	
				dprintf(stderr,"s: ring %d\n",ring);
				set_timeout(50);		/* max time to wait for next ring */
				break;
			default:
				dprintf(stderr,"s: expected r or R got 0x%x [%c]\n",*line,*line);
				break;
			}
		}
	alarm(0);
	if (ring == rings) {		/* avoid false hangup during off-hook */
		int on = loop_off(phone_fd);
		off_hook(phone_fd);
		sleep(2);
		if (on)
			loop_on(phone_fd);
		}
	return(ring);
	}

/* play a file, wait for a tone or timeout */

char
play_file(sound_fd,audio_fd,phone_fd,list,tm,samples)
int sound_fd;		/* fd of sound file */
int audio_fd;		/* fd of audio file */
int phone_fd;		/* fd of phone_fd */
char *list;			/* list of termination contidions */
int tm;				/* timeout (10th seconds) after message ends */
int samples;		/* max number of samples (0-no limit) */
	{
	int done=0;			/* done writing file to kernal Q */
	int hunk = HUNK;	/* maximum size of write chunk */
	int n;

	if (samples<=0)
		samples = MAX_PLAY;

	if (index(list,C_IFLUSH))
		phone_flush(phone_fd);
	audio_reset(audio_fd);
	set_timeout(0);
	dprintf(stderr,"s: Playing file till [%s] or %d/10\n",list,tm);
	while(is_timeout()==0) {
		if (!done && (n = read(sound_fd,buff,Min(samples,hunk)))>0) {
			audio_write(audio_fd,buff,n);
			samples -= n;
			hunk = audio_writeq(audio_fd)>2000?HUNK/2:HUNK;	/* experimental */
			}
		else if (!done) {								/* done reading audio */
			dprintf(stderr,"s: Play done writing file (flushing)\n");
			done++;
			set_timeout(1+tm+audio_writeq(audio_fd)/800);
			}
		n = phone_read(phone_fd,line,sizeof(line),1);
		if (n>0 && index(list,*line)) {
			alarm(0);
			if (!index(list,C_OFLUSH)) {
				play_stop(audio_fd);
				audio_flush(audio_fd,1);
				play_start(audio_fd);
				}
			dprintf(stderr,"s: Play: done returning %c\n",*line);
			return(*line);
			}
		}
	alarm(0);
	dprintf(stderr,"s: Play: Timeout\n");
	return(C_TIMEOUT);
	}

/* wait for a touch tone for a certain length of time */
	
char
fetch_tone(phone_fd,list,tm)
int phone_fd;				/* phone descriptor */
char *list;					/* termination conditions */
int tm;						/* timeout for gathering digits 10th seconds */
	{
	dprintf(stderr,"s: looking for tones [%s]\n",list);
	if (tm) set_timeout(tm);
	while(!is_timeout()) {
		phone_read(phone_fd,line,sizeof(line),0);
		if (index(list,*line)) {
			if (tm) alarm(0);
			return(*line);
			}
		}
	if (tm) alarm(0);
	return(C_TIMEOUT);
	}

/* wait for some touch tones for a certain length of time
 * '*' - start over
 * '#' - done
 */
	
int
fetch_tones(phone_fd,digits,max,list,tm)
int phone_fd;				/* phone descriptor */
char *digits;					/* place to put list */
int max;						/* max # of digits */
char *list;					/* valid codes */
int tm;						/* timeout for gathering digits 10th seconds */
	{
	int i=0;
	char c='x';

	set_timeout(tm);
	while(i < max && c!='#') {
		c = fetch_tone(phone_fd,list,0);
		if (is_timeout())
			break;
		if (c=='*') {		/* restart */
			i = 0;
			continue;
			}
		digits[i++] = c;
		}
	digits[i]='\0';
	dprintf(stderr,"User keyed: %s\n",digits);
	return(i);
	}

/* these should be settable parameters */

#define THRES		75
#define MIN			(HUNK/10)

/* record a file 'till something happens */

char
get_file(sound_fd,audio_fd,phone_fd,list,tm,silence)
int sound_fd;		/* fd of sound file */
int audio_fd;		/* fd of audio file */
int phone_fd;		/* fd of phone_fd */
char *list;			/* list of termination contidions */
int tm;				/* overall timeout (10th seconds) */
int silence;		/* 1/10th seconds of silence */
	{
	int count = 0;	/* # of hunks read (for dial tone detection */
	int quiet = index(list,C_QUIET) ? 1 : 0;
	int dial = index(list,C_DIAL) ? 1 : 0;
	int rd;		/* bytes of audio read */
	int n;
	char result = '\0';

	silence *= 800;				/* convert 10th seconds to samples */
	phone_flush(phone_fd);
	record_start(audio_fd);
	set_timeout(tm);
	while(!is_timeout() && (rd = read(audio_fd,buff,HUNK))>0) {
		if (rd != write(sound_fd,buff,rd)) {
			result='?';
			break;
			}
		n = phone_read(phone_fd,line,3,1);
		if (n>0 && (index(list,*line) || *line == 'd')) {
			dprintf(stderr,"s: Get_file: done returning %c\n",*line);
			result = *line;
			break;
			}

		/* do quiet detection */

		if (quiet)  {
			quiet = is_quiet(buff,rd,THRES,MIN) ? quiet+rd : 1;
			}
		if (quiet > silence) {
			dprintf(stderr,"s: Get: got silence\n");
			result = C_QUIET;
			break;
			}

		/* do dial-tone detection - in case quiet detection fails */

		if (dial && count++%10==0 && is_dial(buff,rd)) {
			dprintf(stderr,"s: Get: got dial tone\n");
			result = C_DIAL;
			break;
			}
		}
	alarm(0);
	dprintf(stderr,"s: Get: done flushing file\n");
	if (!index(list,C_OFLUSH) || !is_timeout())	/* keep going */
		record_stop(audio_fd);
	return(is_timeout() ? C_TIMEOUT:result);
	}

/* play an orator file */

char
play_orfile(s,audio_fd,phone_fd,list,tm)
char *s;		/* orator cookie */
int audio_fd;
int phone_fd;
char *list;
int tm;		/* timeout in 10ths */
	{
	int or_fd;
	char c;

	if ((or_fd = or_getfd(s)) <= 0)
		return(0);
	c = play_file(or_fd,audio_fd,phone_fd,list,tm,0);
	while(c!='T' && read(or_fd,buff,HUNK) > 0)
		dprintf(stderr,"s: flushing (or)\n");		/* flush rest of file */
	or_end(s);
	return(c);
	}

/* save orator output to a file */

int
save_orfile(s,fd)
char *s;			/* orator cookie */
int fd;	/* where to save the speech */
	{
	int or_fd;
	int sum = 0;
	int n;

	if ((or_fd = or_getfd(s)) ==0)
		return(0);
	while((n=read(or_fd,buff,HUNK)) > 0) 
		sum += write(fd,buff,n);
	Dprintf(stderr,"S: Saved %d bytes of OR output\n",sum);
	or_end(s);
	return(sum);
	}

void
debug_subs(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"subs_debug in %d\n",n);
	}

/* global debug setting */

#include <ctype.h>

void debug_audio(), debug_db(), debug_or(), debug_phone(), debug_subs();
void debug_dsp(), debug_isdial(), debug_progress(), debug_tone();

void
debug_set()
	{
	char *getenv();
	char *list = getenv("DEBUG");
	char c;

	while (list && (c = *list++)) {
		switch(tolower(c)) {
			case 'a': debug_audio(isupper(c)?2:1); break;
			case 'd': debug_db(isupper(c)?2:1); break;
			case 'o': debug_or(isupper(c)?2:1); break;
			case 'p': debug_phone(isupper(c)?2:1); break;
			case 's': debug_subs(isupper(c)?2:1); break;
			case 'c': debug_dsp(isupper(c)?2:1);
			          debug_isdial(isupper(c)?2:1);
			          debug_progress(isupper(c)?2:1);
			          debug_tone(isupper(c)?2:1); break;
			case '?':
				fprintf(stderr,"Debug flags:\n");
				fprintf(stderr," a:audio\n");
				fprintf(stderr," d:database\n");
				fprintf(stderr," o:orator\n");
				fprintf(stderr," p:phone\n");
				fprintf(stderr," s:call manipulation\n");
				fprintf(stderr," c:call origination\n");
				break;
			default:
				fprintf(stderr,"Invalid debug option, %c\n",c);
			}
		fprintf(stderr,"Debugging set to %s\n",getenv("DEBUG"));
		}
	}
