/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* touch tone decoder stuff (8khz version) */

#include <stdio.h>
#include <math.h>
#include "dsp.h"

/* constants & misc */

#define FILTERS	7					/* number of filters (7 or 8) */
#define NQ			(FILTERS+1)		/* number of Q's */
#define MAXQ		4					/* max elements in each Q (power of 2) */
#define INQ			FILTERS			/* name of input Q */
#define SMPMS		8					/* samples per msec */
#define ON			1					/* on, waiting for off */
#define OFF			0					/* off, waiting for on */

/* adjustable parameters */

#define DRAIN		0.1				/* integrator drain (you pick) */
#define THRES1		0.7				/* theshold for ON detection */
#define THRES2		0.5				/* theshold for OFF detection */
#define SAMPLES	(75*SMPMS)		/* # samples for detection */
#define GAIN		4000.0			/* initial input gain factor */

#define BETWEEN(a,x,b)     ((x)<(a)?a:((x)>(b)?b:x))
#define ABS(x)					((x)<0.0 ? -(x) : (x))

/* band pass filter element */

#define Filter(c,x0,x2,y1,y2) \
	(c)->k*((x0) - (x2)) - ((c)->b1*(y1) + (c)->b2*(y2))

/* Queue management */

#define ADDQ(i,x)		(q[((i)<<2)+cq] = (x))
#define FETCHQ(i,n)	(n+cq >=0 ? q[((i)<<2)+n+cq] : q[((i)<<2)+cq+n+MAXQ])
#define BUMPQ()		(cq = (cq+1)&3)

struct sum {
	float sum;		/* current integrated value */
	int state;		/* current output state */
	int count;	/* current count in "other" state */
	};

/* coefficents for touch tone band pass filters
 * sampling rate 8khz, filter gain: 15
 */

static struct coef coef[] = {
	/*   k     a0   a1    a2        b1        b2      	    f0       gain */
	0.01171212, 1.00000000, 0.00000000, -1.00000000, -1.69773720, 0.98828788,
	0.01171212, 1.00000000, 0.00000000, -1.00000000, -1.63564617, 0.98828788,
	0.01171212, 1.00000000, 0.00000000, -1.00000000, -1.55950066, 0.98828788,
	0.01171212, 1.00000000, 0.00000000, -1.00000000, -1.46954812, 0.98828788,
	0.00585606, 1.00000000, 0.00000000, -1.00000000, -1.15728696, 0.98828788,
	0.00585606, 1.00000000, 0.00000000, -1.00000000, -0.99053541, 0.98828788,
	0.00585606, 1.00000000, 0.00000000, -1.00000000, -0.79394163, 0.98828788,
	0.00585606, 1.00000000, 0.00000000, -1.00000000, -0.56520335, 0.98828788,
#ifdef OLD	
	0.00544448, 1.0, 0.0, -1.0, -0.91137928, 0.98911105,		/* 697 */
	0.00601129, 1.0, 0.0, -1.0, -0.70270001, 0.98797743,		/* 770 */
	0.00664721, 1.0, 0.0, -1.0, -0.45771596, 0.98670558,		/* 852 */
	0.00733651, 1.0, 0.0, -1.0, -0.18373084, 0.98532698,		/* 941 */
	0.00941414, 1.0, 0.0, -1.0,  0.63879101, 0.98117173,		/* 1209 */
	0.01041511, 1.0, 0.0, -1.0,  0.99675582, 0.97916978,		/* 1336 */
	0.01149084, 1.0, 0.0, -1.0,  1.34655574, 0.97701832,		/* 1477 */
	0.01264085, 1.0, 0.0, -1.0,  1.65557475, 0.97471830,		/* 1633 */
#endif
	};

static float q[NQ*MAXQ];			/* current Q values */
static struct sum sums[FILTERS];		/* place to hold sums */
static int cq = 0;						/* current index into Q's */
static int get_digit();
char *getenv();

#define dprintf	if(debug)fprintf
static int debug = 0;

extern unsigned char  Ub2i[8192];
extern unsigned short I2ub[256];
extern unsigned char  *Sb2i;
extern short I2sb[256];

static int first = 1;
static int state = 0;		/* global filter state */
static int digit = 0;		/* ?? */

#define	B13ZERO	0x1000		/* zero offset in unsigned 13-bit tables */

int
tones(in,count,result)
register char *in;							/* audio input data */
int count;							/* # of samples */
int *result;						/* put result here */
	{
	char *end = in + count;	
	register int c;				/* current u-law input sample */
	register int i;				/* filter number */
	register float x;				/* intermediate value */
	register float x0,x2;		/* input values */

	float thres = THRES1;
	float gain = GAIN;
	float drain = DRAIN;
	int minsamp = SAMPLES;

	/* initialization */

	if (first) {
		gain = getenv("GAIN") ? atof(getenv("GAIN")) : GAIN;
		drain = getenv("DRAIN") ? atof(getenv("DRAIN")) : DRAIN;
		}
	if (count == 0 || first) {
		first = state = digit = 0;
		for(i=0;i<FILTERS;i++) {
			sums[i].sum = 0.0;
			sums[i].state = OFF;
			sums[i].count = 0;
			}	
		for(i=0;i<NQ*MAXQ;i++)
			q[i] = 0.0;
		}

	while (in < end) {
		x = I2sb[(int)*in++]/gain;	/* scale to +- 1 */
		ADDQ(INQ,x);			/* feed the pipeline */
		x0 = FETCHQ(INQ,0);
		x2 = FETCHQ(INQ,-2);

		/* do the filters */

		for(i=0;i<FILTERS;i++) {

			/* do the filter stage */

			x = Filter((coef+i), x0, x2, FETCHQ(i,-1), FETCHQ(i,-2));
			ADDQ(i,x);

			/* compute the thresholds */

			x = ABS(x);
			x += sums[i].sum - drain;
			sums[i].sum = BETWEEN(0.0,x,1.0);
			switch (sums[i].state) {
				case OFF:		/* look for as hit */
					if (x > thres)
						sums[i].count++;
					else
						sums[i].count = 0;
					break;	
				case ON:		/* wait for decay  */
					if (x < THRES2)
						sums[i].count++;
					else
						sums[i].count = 0;
					break;	
				}

			/* got a tone transition */

			if (sums[i].count > minsamp) {
				sums[i].count = 0;
				sums[i].state = 1-sums[i].state;
				if (sums[i].state == ON)
					state |= (1<<i);
				else
					state &= ~(1<<i);

				c = get_digit(state);
				dprintf(stderr,"tones: state %x got %d\n", state,c);
				if (c>0 && c != digit) {
					digit = *result = c;
					dprintf(stderr,"tones: returning %d\n",digit);
					return(in-end+count);
					}
				}
			}
		BUMPQ();
		}
	*result = 0;
	digit = 0;
	return(in-end+count);
	}

/* convert state bit mask into tones (should be a look-up table) */

static int
get_digit(code)
int code;
	{
	switch(code) {
		case 0x11: return(1);
		case 0x21: return(2);
		case 0x41: return(3);
		case 0x12: return(4);
		case 0x22: return(5);
		case 0x42: return(6);
		case 0x14: return(7);
		case 0x24: return(8);
		case 0x44: return(9);
		case 0x18: return(10);
		case 0x28: return(11);
		case 0x48: return(12);
		case 0:     return(0);
		default:	   return(0);
		}
	}

void
debug_tone(n)
int n;
	{
	if (debug = n)
		fprintf(stderr,"tones_debug in %d\n",n);
	}
