/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* experimental PhoneStation gdbm interface (S. A. Uhler  1/93) */

/* Preliminary Command interface
 *  set handle [dbm open path how]
 *    how: read, write, create (implies write).  Defaults to read
 *  dbm insert handle key [value]
 *  dbm delete handle key
 *  dbm replace handle key value
 *  dbm fetch handle key
 *	 dbm next handle [key]  <no key gives first key>
 *  dbm close handle
 */

#include <malloc.h>
#include "tcl.h"
#include "defs.h"
#include "gdbm.h"

#define MAXGDBMS	10	/* max # of dbm files opened at once */
#define PREFIX		"GDB."

/* legal db options (order must match up with case statement below (yuk) */

struct cmds {
	char *name;			/* command name */
	int min,max;		/* valid # of arguments */
	int write;			/* true if needs write permission */
	};

struct handles {
	GDBM_FILE handle;		/* gdbm handle */
	int write;				/* true iff write flag on open */
	};

static struct cmds cmds[] = {
	"open",		3, 4, 0,
	"insert", 	4, 5, 1,
	"delete",	4, 4, 1,
	"replace",	5, 5, 1,
	"fetch",		4, 4, 0,
	"next",		3, 4, 0,
	"close",		3, 3, 0,
	NULL, 		0, 0, 0,
	};

char *modes = "rwc";				/* valid open modes, read, write, or create */
struct handles handles[MAXGDBMS];
char *index();
static datum make_datum(char *s);

int
gdbm(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	int how;				/* open mode */
	datum result;		/* result from fetch */
	GDBM_FILE handle=NULL;	/* current gdbm handle */
	int indx=0;			/* index into handle table */
	int i;				/* command index and return code */

   argc = parse_args(argv);
   if (argc < 3) {
      Tcl_AppendResult(interp,*argv,": not enough parameters\n",NULL);
      return(TCL_ERROR);
      }
   dprintf("T","In on as %s\n",argv[1],argc);

	/* check options */

	for(i=0;cmds[i].name;i++)
		if (strcmp(argv[1],cmds[i].name)==0)
			break;

	if (!cmds[i].name) {
      Tcl_AppendResult(interp,*argv,": invalid option\n",NULL);
      return(TCL_ERROR);
      }

	if (argc<cmds[i].min || argc>cmds[i].max) {
      Tcl_AppendResult(interp,*argv,": ",argv[1]," wrong # of parameters\n",
					NULL);
      return(TCL_ERROR);
      }

	if (i && strncmp(PREFIX,argv[2],strlen(PREFIX))!=0) {
      Tcl_AppendResult(interp,*argv,": invalid database handle\n",NULL);
      return(TCL_ERROR);
      }

	if (i) {
		indx = atoi(argv[2]+strlen(PREFIX));
		if (!(handle = handles[indx].handle)) {
			Tcl_AppendResult(interp,*argv,": invalid database handle\n",NULL);
			return(TCL_ERROR);
			}
		if (handles[indx].write==0 && cmds[i].write==1) {
			Tcl_AppendResult(interp,*argv,
					": invalid option on readonly database\n",NULL);
			return(TCL_ERROR);
			}
		}

	switch(i) {
		case 0:			/* open */
			if (argc>3 && index(modes,*argv[3]))
				how = index(modes,*argv[3])-modes;
			else
				how = 0;
			handle = gdbm_open(argv[2],0,how,0640,NULL);
			if (!handle) {
				Tcl_AppendResult(interp,*argv,":database won't open\n",NULL);
				return(TCL_ERROR);
				}
			for(i=0;i<MAXGDBMS;i++)
				if (!handles[i].handle) {		/* install handle in handle list */
					handles[i].handle = handle;
					handles[i].write = (how > 0);
					Tcl_SetResult(interp,sprintf(line,"%s%d",PREFIX,i),TCL_VOLATILE);
					return(TCL_OK);
					}
			gdbm_close(handle);
			Tcl_AppendResult(interp,*argv,":too many open databases\n",NULL);
			return(TCL_ERROR);
			break;	
		case 1:		/* insert */
			i = gdbm_store(handle,make_datum(argv[3]),make_datum(argv[4]),0);
			dprintf("G","dbm insert %s: %s, code %d\n",argv[3],argv[4],i);
			Tcl_SetResult(interp,i?"1":"0",TCL_VOLATILE);
			break;
		case 2:		/* delete */
			i = gdbm_delete(handle,make_datum(argv[3]));
			dprintf("G","dbm delete %s, code %d\n",argv[3],i);
			Tcl_SetResult(interp,i?"1":"0",TCL_VOLATILE);
			break;
		case 3:		/* replace */
			i = gdbm_store(handle,make_datum(argv[3]),make_datum(argv[4]),1);
			dprintf("G","dbm replace %s: %s, code %d\n",argv[3],argv[4],i);
			Tcl_SetResult(interp,sprintf(line,"%d",i),TCL_VOLATILE);
			break;
		case 4:		/* fetch */
			result = gdbm_fetch(handle,make_datum(argv[3]));
			dprintf("G","dbm fetch %s = %s\n",argv[3],
					result.dptr?result.dptr:"(NULL)");
			if (result.dptr)
				Tcl_SetResult(interp,result.dptr,TCL_DYNAMIC);
			else
				Tcl_SetResult(interp,"",TCL_VOLATILE);
			break;
		case 5:		/* next */
			if (argc > 3) 		/* db next */
				result = gdbm_nextkey(handle,make_datum(argv[3]));
			else 					/* db first */
				result = gdbm_firstkey(handle);
			if (result.dptr)
				Tcl_SetResult(interp,result.dptr,TCL_DYNAMIC);
			else
				Tcl_SetResult(interp,"",TCL_VOLATILE);
			break;
		case 6:		/* close */
			gdbm_close(handle);
			handles[indx].handle = NULL;
			break;
		}
	return(TCL_OK);
	}

/* make a "datum" entry from a string */

static datum
make_datum(s)
char *s;
	{
	static datum result;

	result.dptr = s;
	result.dsize = strlen(s)+1;
	return(result);
	}

/* close all open databases */

int
gdbm_closeall()
	{
	int i;

	for(i=0;i<MAXGDBMS;i++)
		if (handles[i].handle)
			gdbm_close(handles[i].handle);
	return(1);
	}
