/****************************************************************
Copyright (c) 1993 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material
for any purpose and without fee is hereby granted, provided
that the above copyright notice and this permission notice
appear in all copies, and that the name of Bellcore not be
used in advertising or publicity pertaining to this
material without the specific, prior written permission
of an authorized representative of Bellcore.  BELLCORE
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
****************************************************************/

/* play a file until an event happens (this is the play command) */

#include <stdio.h>
#include <sys/signal.h>
#include <sys/file.h>
#include <sys/time.h>
#include <fcntl.h>
#include <malloc.h>
#include "tcl.h"
#include "defs.h"

/* usage
 *
 *	
 *	play {files} until {regex} unless {expr}  [options]
 *	play {files} until {regex} [options]
 *	play {files} {regex} {expr}  [options]
 *	play {files} {regex} [options]
 *	play {file}			< this is currently a special case >
 *	
 *	options
 *		timelimit=nnn		Set a max time limit, after last file is done playing
 *		ignore=<list>		These touch tones are always ignored (just a list)
 *		maxtones=nnn		maximum # of tones to gather
 *    beep=1				always beep when play is done 
 */

/* set timeout to start when audio Q drains */

#define Set_timer() \
		set_timeout(1+timelimit+audio_writeq(S(audio_fd))/800)

#define BETWEEN(a,x,b)      (x)<(a)?a:((x)>(b)?b:x)

#define MAXHUNK	24000		/* max hi-water mark size */
#define HUNK		800		/* default size (1/10 second) */
#define MINHUNK	HUNK			/* minimum # samples to read, even if Q too big */
#define UNLESS		"unless"	/* name of unless variable array */

/* play/unless variable exchange */

#ifndef OLD_PLAY
#define SET(interp,name,value,flags)	\
		Tcl_SetVar2(interp,UNLESS,name,value,flags)
#define GET(interp,name,flags) \
		Tcl_GetVar2(interp,UNLESS,name,flags)
#define TRACE(interp,name,flags,func,var) \
		Tcl_TraceVar2(interp,UNLESS,name,flags,func,(ClientData) var)
#define UNTRACE(interp,name,flags,func,var) \
		Tcl_UntraceVar2(interp,UNLESS,name,flags,func,(ClientData) var)
#else
#define SET(interp,name,value,flags)	\
		Tcl_SetVar(interp,name,value,flags)
#define GET(interp,name,flags) \
		Tcl_GetVar(interp,name,flags)
#define TRACE(interp,name,flags,func,var) \
			Tcl_TraceVar(interp,name, flags, func,(ClientData) var)
#define UNTRACE(interp,name,flags,func,var) \
			Tcl_UntraceVar(interp,name,flags,func,(ClientData) var)
#endif

struct var {					/* variables x-ferred between C and TCL */
	char *name;					/* name of variable */
	char *addr;					/* pointer to variable containing it */
	char *new;					/* new value */
	char flags;					/* type, NUM or CHAR (see below) */
	char special;				/* special handling required */
	char recurse;				/* recursion level this applies for */
	};

struct fc {						/* file control structure */
	int fd;						/* file descriptor for current file (>0) */
	char **file_list;			/* current list of files to play */
	int file;					/* current file in process */
	int filec;					/* # of files in file_list */
	int is_or;					/* true if current file is from orator */
	};

char *index();
char *regcomp();
extern char *tclRegexpError;

static int get_reason(char *s); /* !!proto */
static int next_file(struct fc *fc, struct stuff *stuff); /* !!proto */
static struct fc *setup_files(struct fc *fc, Tcl_Interp *interp, char *list, int start); /* !!proto */
static void close_file(struct fc *fc, struct stuff *stuff); /* !!proto */
static char *do_trace(ClientData data, Tcl_Interp *interp, char *name1, char *name2, int flags); /* !!proto */

/* play might be called recursively, so watch what goes here */

static char buff[MAXHUNK];		/* audio buffer */
static char line[3];			/* phone input line */
static int recurse=0;		/* Current recursion level */
static char *reasons[] = {
	"tfdx", "Tfdx", "tFdx", "TFdx", "tfDx", "TfDx", "tFDx", "TFDx",
	"tfdX", "TfdX", "tFdX", "TFdX", "tfDX", "TfDX", "tFDX", "TFDX",
	};

/* types variables exchanged between play and unless (preliminary) */

#define CHAR		2		/* string variable */
#define NUM			4		/* numeric variable */
#define PNTR		8		/* pointer to value (broken) */
#define READ		16		/* variable was already read */
#define WRITE		32		/* variable was written to */
#define LOCK		64		/* in trace already */

/* variables that need special post processing */

#define S_LOW_WATER	1	/* minimum audio output Q length */
#define S_FILE			2	/* current playing file # */
#define S_TONE			3	/* current tone */
#define S_FILE_POS	4	/* current file lseek() position */
#define S_FILES		5	/* list of files to play */
#define S_RDONLY		6	/* (read only) in another unless clause */
#define S_TIMELIMIT	7	/* change the timelimit */
#define S_REASON		8	/* reason for running unless changed */

/* Reasons for running unless expression */

#define UN_TONE		1	/* got a tone */
#define UN_NEXT		2	/* next file */
#define UN_DONE		4	/* done with files */
#define UN_TIME		8	/* time-out */

/* process variables (this requires ansi style string-ification) */

#define MAXVARS	20		/* Max # of variables exchanged */
#define start_vars()	(i=0)
#define end_vars()	(vars[i].name = NULL)
#ifdef CC_BOTCH
#define set_var(Name,Type,Special) ( \
	vars[i].name = "Name", \
	vars[i].addr = (char *) &Name, \
	vars[i].flags = Type, \
	vars[i].new = "", \
	vars[i].recurse = (char) recurse, \
	vars[i].special = Special, i++)
#else
#define set_var(Name,Type,Special) ( \
	vars[i].name = #Name, \
	vars[i].addr = (char *) &Name, \
	vars[i].flags = Type, \
	vars[i].new = "", \
	vars[i].recurse = (char) recurse, \
	vars[i].special = Special, i++)
#endif
	
/* play files and gather touch tones */

int
play(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char *argv[];
	{
	struct var *var;
	struct stuff *stuff = (struct stuff *) clientData;
	struct fc *fc=NULL;	/* cookie for file manipulation routines */
	char *name = *argv;	/* function name */
	char *regex=NULL;		/* "until" regular expression */
	char *expr;				/* "unless" expression to evaluate (if any) */
	int done = 0;			/* true iff done playing all files */
	char *bad;				/* pointer to bad command in TCL unless expr */
	char *ignore;			/* tones to always ignore */
	char *reg_list;		/* pointer to compiled regex for "until" */
	int error = 0;			/* error in arg list */
	int new=0;				/* change happened to file list or index */
	int n;					/* # of chars read/written */
	int reason=0;			/* run unless expr for these changes */
	int do_unless = 0;	/* need to run unless expression */
	int type_ahead;		/* allow type_ahead */
	int unless_count=0;	/* # of times unless is called */
	int file;				/* current file # */
	int filec;				/* current # of files */
	int maxtones;			/* max number of tones to fetch */
	int sound_fd;			/* fd of currently playing file */
	int timelimit;			/* max time after last files finishes (10th sec) */
	int low_water = HUNK;	/* low water mark for sound Q */
	int file_pos;			/* current file position */
	int beep = 0;			/* beep when done playing */
	int digits = 0;		/* # of tones */
	int leave;		/* how much time to leave for "just play" */
	int nofile_ok;			/* OK if file doesn't exist */
	char tone[2];			/* just received tone */
	char why[6];			/* reason that unless was called */
	char tones[50];		/* place to hold tones */
	struct var vars[MAXVARS];		/* array of exported variables */
	char file_name[100];	/* current file name */
	char files[4096];			/* list of files to play */

	/* check arguments */

	argc = parse_args(argv);
	dprintf("T","In play (%d) [%.32s...]\n",recurse,argv[1]);
   set_timeout(0);

	switch (argc) {
		case 6:
			if (strcmp(argv[2],"until") || strcmp(argv[4],"unless"))
				error++;
			else {
				regex = argv[3];
				expr = argv[5];
				}
			break;
		case 4:
			if (strcmp(argv[2],"until")) {
				regex = argv[2];
				expr = argv[3];
				}
			else {
				regex = argv[3];
				expr = NULL;
				}
			break;
		case 3:
			regex = argv[2];
			expr = NULL;
			break;
		case 2:							/* just play a file, don't look for tones */
			leave = atoi(get_value(interp,"leave"));
			nofile_ok =(1 == is_true(get_value(interp,"nofile_ok")));
			leave = BETWEEN(HUNK,leave,MAXHUNK);
			return(just_play(interp,stuff,argv[1],leave,nofile_ok));
			break;
		default:
			error++;
			break;
		}

	if (error) {
		Tcl_AppendResult(interp,"Usage: ",name,
				" <files...> until <regexp> unless {cmds} [options]\n",NULL);
		return(TCL_ERROR);
		}

	/* get regular expressions */

	if ((reg_list = regcomp(regex)) == NULL) {
		Tcl_AppendResult(interp,"couldn't compile <regex>: ",
					tclRegexpError,"\n",NULL);
		return(TCL_ERROR);
		}

	/* setup files to play */

   if (!recurse)		/* reset audio only on top-level call */
		audio_reset(S(audio_fd));

	strcpy(files,argv[1]);
	dprintf("p","Starting off playing files [%s] at 0x%x\n",files,files);
	if (!(fc=setup_files(fc,interp,files,0))) {
		Tcl_AppendResult(interp,name,": Can't parse files ",argv[1],"\n",NULL);
		return(TCL_ERROR);
		}

	/* get keyword values */

	tone[1]='\0';
	ignore = get_value(interp,"ignore");
	type_ahead = atoi(get_value(interp,"type_ahead"));
	nofile_ok =  (1==is_true(get_value(interp,"nofile_ok")));
	maxtones = atoi(get_value(interp,"maxtones"));
	timelimit = atoi(get_value(interp,"timelimit"));
	beep = (1 == is_true(get_value(interp,"beep")));
	reason = get_reason(get_value(interp,"reason"));
	low_water = atoi(get_value(interp,"leave"));
	if (low_water ==0)
		low_water = HUNK;
	else
		low_water = BETWEEN(MINHUNK,low_water,MAXHUNK);
	bzero(tones,sizeof(tones));
	strcpy(tones,get_value(interp,"start_tones"));
	recurse++;

	/* set up variables to export to TCL "unless" clause */

	if (expr) {
		register int i;

		start_vars();
		set_var(reason,CHAR,S_REASON);		/* the reason for unless changed */
		set_var(expr,CHAR,S_RDONLY);			/* the unless expression (broken) */
		set_var(done,NUM,S_RDONLY);			/* files are done playing */
		set_var(do_unless,NUM,S_RDONLY);		/* reasons for calling do_unless */
		set_var(maxtones,NUM,0);				/* Max # of tones permitted */
		set_var(timelimit,NUM,S_TIMELIMIT);	/* timeout */
		set_var(low_water,NUM,S_LOW_WATER);	/* max audio Q size */
		set_var(file,NUM,S_FILE);				/* Current file # */
		set_var(files,CHAR,S_FILES);			/* current files */
		set_var(tone,CHAR,0);					/* Current Tone */
		set_var(tones,CHAR,0);					/* list of tones */
		set_var(digits,NUM,0);					/* # of tones so far */
		set_var(unless_count,NUM,0);			/* # of times unless was run */
		set_var(ignore,CHAR,0);					/* list of tones to always ignore */
		set_var(why,CHAR,S_RDONLY);			/* why unless was called */
		set_var(file_name,CHAR,S_RDONLY);	/* current file name */
		set_var(file_pos,NUM,S_FILE_POS);	/* current file position */
		end_vars();
		set_trace(interp,vars);
		}

	if (!type_ahead)
		phone_flush(S(phone_fd));

	/* setup first file (if any) */

	if ((sound_fd = next_file(fc,stuff)) <= 0) {
		done++;
		if (!nofile_ok) {
			Tcl_AppendResult(interp,"Couldn't open file: ",name," to play\n",NULL);
			return(TCL_ERROR);
			}
		Set_timer();
		}

	/* main loop */

	while(!S(hangup)) {

		if (is_timeout()) {
			*tone = 'T';
			if ((do_unless = (reason&UN_TIME)))
				Set_timer();		/* reset the timer, run the unless clause */
			else
				break;				/* all done! */
			}

		if (!done) {									/* still playing files */
			int samples = audio_writeq(S(audio_fd)) - low_water;

			/* wait for audio to drain to low water mark */

			ssleep(samples);

			dprintf("P","reading %d from %d\n",HUNK,sound_fd);
			if (sound_fd>0 && (n = read(sound_fd,buff,HUNK))>0) {
				dprintf("P"," got %d samples of audio\n",n);
				audio_write(S(audio_fd),buff,n);
				dprintf("P"," wrote %d samples of audio\n",n);
				}
			else if ((sound_fd = next_file(fc,stuff)) <= 0) {	/*files done*/
				done++;
				if (beep) {
					audio_beep(S(audio_fd),2,0);
					dprintf("P","beeping\n");
					}
				dprintf("p","Done playing files (none left) [%s]\n",files);
				do_unless |= (reason&UN_DONE);
				Set_timer();
				}
			else {										/* re-try with next file */
				dprintf("p","Now playing file %d/%d\n",fc->file,fc->filec);
				if (reason&UN_NEXT)
					do_unless |= UN_NEXT;
				else
					continue;		/* compatibility with previous versions */
				}
			}
		n = phone_read2(S(phone_fd),line,sizeof(line));
		if (n>0 && index(ignore,*line)==0) {					/* handle tone */
			*tone = *line;
			dprintf("p","Got tone [%c] -> [%s]\n",*tone,tones);
			do_unless |= (reason&UN_TONE);
			}
		else if (n>0)
			dprintf("p","Ignoring tone [%c]\n",*tone);

		/* run unless expression */

		if (expr && do_unless) {

			/* pre-process variables */

			file = fc->file;
			filec = fc->filec;
			file_pos = tell(sound_fd);
			if (file<filec)
				strcpy(file_name,fc->file_list[file]);
			else
				*file_name = '\0';
			strcpy(why,reasons[do_unless]);
			digits = strlen(tones);
			unless_count++;

			/* run the unless expression */

			dprintf("p","Running unless clause\n");
			n = Tcl_Eval(interp,expr,0,&bad);
			dprintf("p","expr returned %s\n",print_code(n));

			/* handle any variables changed by "unless" */

			dprintf("P","Handle unless changes:\n");
			for(var=vars;var->name;var++) {
				dprintf("P","***  var %s 0x%x (now %s) special=%d\n",
							var->name, var->flags, var->new, var->special);
				if (!(var->flags&WRITE)) {
					var->flags &= NUM|CHAR|PNTR;
					continue;
					}
				dprintf("P","   changed\n");
				switch(var->special) {
					case 0:				/* no special handling needed */
						if (var->flags & NUM)
							*((int *)var->addr) = atoi(var->new);
						else
							strcpy(var->addr,var->new);	/* no overflow checking */
						break;
					case S_RDONLY:		/* read only variables */
						break;
					case S_FILE_POS:
						if (sound_fd > 0) {
							dprintf("p","Changing file_pos: %d to %d\n",
									file_pos, atoi(var->new));	
							lseek(sound_fd,atoi(var->new),0);
							}
						break;
					case S_TIMELIMIT:
						timelimit = atoi(var->new);
						if (is_timer())	/* reset timelimit if timer is set */
							Set_timer();
						break;
					case S_TONE:
						dprintf("p","Changing tone: %c to %c\n",*tone, *var->new);
						*tone = *var->new;
						break;
					case S_LOW_WATER:
						n = atoi(var->new);
						n = BETWEEN(800,n,MAXHUNK);
						low_water = n;
						break;
					case S_FILES:
						dprintf("p","Changing files: [%s] to [%s]\n", files, var->new);
						strcpy(files, var->new);		/* NO length checking (bug) */
						new++;
						break;
					case S_FILE:
						dprintf("p","Changing file: [%d] to [%d]\n",
										file,atoi(var->new));
						file = atoi(var->new)-1;	/* '-1' ??? */
						if (file<0) file = 0;
						new++;
						break;
					case S_REASON:		/* read only variables */
						reason = get_reason(var->new);
						break;
					default:
						fprintf(stderr,"Unknown special action for %s: %d\n",
								var->name,var->special);
					}
				var->flags &= NUM|CHAR;
				}

			/* file name or number has changed */

			if (new) {
				fc = setup_files(fc,interp,files,file);
				new = done;					/* temporarily save previous value */
				done =  ((sound_fd = next_file(fc,stuff)) <= 0);
				if (done != new)
					dprintf("p","Play status was %s now %s\n",
									new?"ON":"OFF", done?"ON":"OFF");
				if (done && !is_timer()){	/* set new timeout */
					Set_timer();
					}
				else if (!done && is_timer)	/* clear timer */
					set_timeout(0);
				new = 0;
				}

			/* handle exit codes */

			dprintf("p","UNLESS returned %d\n",n);
			if (n==TCL_BREAK || n==TCL_UNWIND)	/* finish immediately */
				break;
			else if (n==TCL_ERROR) { /* error running <expr> */
				error_cmd(interp,expr,bad);
				Tcl_AppendResult(interp,argv[0],": error running unless","\n",NULL);
				fprintf(stderr,"Play unless expr: %s\n",interp->result);
				}
			}		/* end of unless processing */

		/* check termination conditions */

		if (do_unless==UN_TONE || (do_unless && expr)) {
			strcat(tones,tone);
			if (maxtones<1 || maxtones > sizeof(tones))	/* wrong spot */
				maxtones = sizeof(tones);
			if (strlen(tones) >= maxtones || regexec(reg_list, tones)) {
				dprintf("p","%s matches %s (or >= %d), done\n",
						regex,tones,maxtones);
				break;
				}
			dprintf("p","No match, looking for another tone\n");
			if (n!=TCL_CONTINUE && do_unless&UN_TONE) {
				done++;				/* stop playing files */
				Set_timer();
				dprintf("p","Got tone, done playing files\n");
				}
			}
		*tone = '\0';
		do_unless = 0;
		}
	if (is_timeout()) {
		dprintf("p","Timeout in play\n");
		strcat(tones,"T");
		}
	set_timeout(0);
	Tcl_SetResult(interp,tones,TCL_VOLATILE);
	if (expr)
		reset_trace(interp,vars);
	Tcl_SetVar(interp,"tones",tones,0);			/* for convenience */
	Tcl_SetVar(interp,"tone",tones+strlen(tones)-1,0);
	close_file(fc,stuff);
	if (reg_list) free(reg_list);
	recurse--;
	if (S(hangup)) {
		dprintf("p","Play level %d returning UNWIND\n",recurse);
		return(TCL_UNWIND);
		}
	else {
		dprintf("p","Play level %d returning OK\n",recurse);
		return(TCL_OK);
		}
	}

/********************************************************************/

/* manipulate lists of files for play */

/* setup a list of files */

static struct fc *
setup_files(prev,interp,list,start)
struct fc *prev;							/* file control block */
Tcl_Interp *interp;	/* current TCL interpreter */
char *list;				/* a list of file names */
int start;				/* file to start with */
	{
	struct fc *result;		/* new file structure */

	/* clean up old stuff */

	dprintf("p","In setup files\n");
	if (prev) {
		result = prev;
		if (result->file_list)
			free(result->file_list);
			result->file_list = NULL;
		}
	else if (!(result = (struct fc *) malloc(sizeof(struct fc))))
		return(NULL);

	/* extract files */

	result->is_or = 0;
	result->file = start;
	result->fd = 0;

	if (Tcl_SplitList(interp,list,&(result->filec),
											&(result->file_list)) != TCL_OK) {
		free(result);
		return(NULL);
		}

	return(result);
	}

/* fetch the next file in the list, dispose of the old one properly  */

static int
next_file(fc,stuff)
struct fc *fc;		/* file control block */
struct stuff *stuff;
	{
	if (fc->fd>0 && fc->is_or)
		or_end(S(or));
	else if (fc->fd>0)
		close(fc->fd);
	fc->fd = 0;

	dprintf("p","  Getting file [%s]\n",fc->file_list[fc->file]);

	while(fc->fd<=0 && fc->file < fc->filec) {
		if (!strcmp(fc->file_list[fc->file],"#")) {
			fc->fd = or_getfd(S(or));
			fc->is_or=1;
			}
		else {
			fc->fd = open(fc->file_list[fc->file],0);
			fc->is_or=0;
			}
		fc->file++;
		}
	dprintf(".","next_file returning %d\n",fc->fd);
	return(fc->fd);
	}

/* done with a file */

static void
close_file(fc,stuff)
struct fc *fc;							/* file control block */
struct stuff *stuff;
	{
	dprintf("p","In close file\n");
	if (fc->file_list)
		free(fc->file_list);
	if (fc->fd>0 && fc->is_or) {
		while(read(fc->fd,buff,sizeof(buff))>0)	/* finish it all */
			;
		or_end(S(or));
		}
	else if (fc->fd>0)
		close(fc->fd);
	free(fc);
	}

/* sleep for a given number of samples */

int
ssleep(samples)
int samples;		/* number of samples to sleep */
	{
	struct timeval time; 
	dprintf("p","Sleepying %d samples\n",samples);
	if (samples >0) { 
      time.tv_sec = samples/8000; 
      time.tv_usec = (samples%8000)*1000/8; 
      select(0,0,0,0,&time); 
		}
	return(1);
	}

/* special case of "play file" that is known to work in an unless clause */

int
just_play(interp,stuff,name,left,ok)
Tcl_Interp *interp; 
struct stuff *stuff;
char *name;						/* name of file to play */
int left;						/* bytes to leave in audio Q */
int ok;							/* ok no file error */
	{
	int n;
	int fd = open(name,0);

	dprintf("T","Just playing %s\n",name);
	if (fd < 0) {
		if (ok)
			return(TCL_OK);
		else {
			Tcl_AppendResult(interp,"Couldn't open file: ",name," to play\n",NULL);
			return(TCL_ERROR);
			}
		}

	while((n=read(fd,buff,MAXHUNK)) > 0)
		audio_write(S(audio_fd),buff,n);
	close(fd);
	
	/* wait 'till almost done playing */

	n = audio_writeq(S(audio_fd));
	dprintf(".","%d left in Q (sleeping %d) \n",n,n-left);
	ssleep(n-left);	/* sleep for "n" samples */
	return(TCL_OK);
	}

/* Set traces on TCL variables (re-do this when OLD_PLAY goes away) */

int
set_trace(interp,var)
Tcl_Interp *interp; 
struct var *var;
	{
	int n;

	for(n=0;var->name;var++,n++) {
		SET(interp,"variables",var->name,TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		TRACE(interp,var->name, TCL_TRACE_READS|TCL_TRACE_WRITES,do_trace,var);
		}
	return(n);
	}

/* remove traces on TCL variables */

int
reset_trace(interp,var)
Tcl_Interp *interp; 
struct var *var;
	{
	dprintf("T","un-setting traces for level %d\n",recurse);
	for(;var->name;var++) {
		UNTRACE(interp,var->name,TCL_TRACE_READS|TCL_TRACE_WRITES, do_trace,var);
		}
	return(1);
	}

/* a read or write reference was made to a "special" variable */

static char  *
do_trace(clientData, interp, name1, name2, flags)
ClientData clientData;
Tcl_Interp *interp;
char *name1, *name2;
int flags;
	{
	struct var *var = (struct var *) clientData;
	char cnvt[12];		/* place to hold num - ascii conversions */

	dprintf("T","In trace for %s(%s)\n",name1,name2);

#ifdef OLD_PLAY
	name2 = name1;
#endif

	if (var->recurse != recurse) {
		dprintf("p","Skip tracefor %s: recurse level %d < %d\n",
				name2,var->recurse, recurse);
		return(NULL);
		}
	if (var->flags&LOCK) {
		dprintf("p","Tracing %s during existing trace!!\n",name2);
		return(NULL);
		}
	var->flags |= LOCK;
	dprintf("pz","Trace on %s is 0x%x\n",name2,flags);

	if (flags&TCL_TRACE_READS && !(var->flags & READ)) {	/* only 1 read ref */
		var->flags |= READ;
		if (var->flags & NUM) {	/* numeric read trace */
			sprintf(cnvt,"%d",*((int *)var->addr));
			dprintf("p","Setting TCL (NUM) [%s] = [%s]\n",var->name, cnvt);
			SET(interp,var->name,cnvt,0);
			}
		else if (var->flags & PNTR) {	/* indirect read trace (broken) */
			dprintf("p","Setting TCL [%s] = [%s] [%s] \n",
					*var->name,var->addr,*var->addr);
			SET(interp,var->name,var->addr,0);
			}
		else  {
			dprintf("pz","Setting TCL [%s] = [%s] (%x) \n",
						var->name,var->addr,var->addr);
			SET(interp,var->name,var->addr,0);
			}
		}
	if (flags & TCL_TRACE_WRITES) {	
		var->flags |= WRITE|READ;
		var->new = GET(interp,var->name,0);
		dprintf("pz","new value for [%s] is [%s]\n",var->name,var->new);
		}
	var->flags &= ~LOCK;
	return(NULL);
	}

/* get reasons for running unless clause *
 *
 * t		tones
 * f		next file
 * d		done
 * x		timeout
 * 
 */

static int
get_reason(s)
char *s;
	{
	register int i=1, result = 0;
	char *p;
	char *list = "tfdx";

	if (!*s)
		return(UN_TONE);
	
	for(i=0;s[i];i++)
		if ((p=index(list,s[i]))) {
			result |= 1<<(p-list);
			}
	return(result);
	}
